using System;
using System.Runtime.InteropServices; /* Provides mappings between C# and native code */

/* This example calls either a Windows or Linux version of the NAG Library based on
 * which OS the example is being executed on*/

namespace NAG
{

    class E04vh
    {

        /* NAG routine declarations copied from flcsdnet64.cs & flcsdnet64_linux.cs*/

        /* Both Windows and Linux routine declarations included as the dll is not 
         * loaded unless the routine is called explicitly.
         * The Linux declaration has an underscore appended to differentiate it*/

        public delegate void E04VHF_USRFUN_DELEGATE(
                                                  ref int STATUS,
                                                  ref int N,
                                                  IntPtr X_rptr,
                                                  ref int NEEDF,
                                                  ref int NF,
                                                  IntPtr F_rptr,
                                                  ref int NEEDG,
                                                  ref int LENG,
                                                  IntPtr G_rptr,
                                                  IntPtr CUSER_chptr,
                                                  IntPtr IUSER_iptr,
                                                  IntPtr RUSER_rptr,
                                                  long CUSER_chptrLength
                                                  );

      [DllImport("NLW6I29DE_nag.dll")] /* Import from DLL, the C# compiler provides a rudimentary check of the signature */
        public static extern void E04VHF(
                                         ref int START, ref int NF, ref int N,
                                         ref int NXNAME, ref int NFNAME, ref double OBJADD,
                                         ref int OBJROW, string PROB, E04VHF_USRFUN_DELEGATE USRFUN,
                                         [In] int[] IAFUN, [In] int[] JAVAR, [In] double[] A,
                                         ref int LENA, ref int NEA, [In] int[] IGFUN,
                                         [In] int[] JGVAR, ref int LENG, ref int NEG,
                                         [In] double[] XLOW, [In] double[] XUPP, [In,Out] byte[,] XNAMES,
                                         [In] double[] FLOW, [In] double[] FUPP, [In,Out] byte[,] FNAMES,
                                         [In,Out] double[] X, [In,Out] int[] XSTATE, [Out] double[] XMUL,
                                         [In,Out] double[] F, [In,Out] int[] FSTATE, [In,Out] double[] FMUL,
                                         ref int NS, ref int NINF, ref double SINF,
                                         [In,Out] byte[,] CW, ref int LENCW, [In,Out] int[] IW,
                                         ref int LENIW, [In,Out] double[] RW, ref int LENRW,
                                         [In,Out] byte[,] CUSER, [In,Out] int[] IUSER, [In,Out] double[] RUSER,
                                         ref int IFAIL,
                                         long PROBLength,
                                         long XNAMESLength,
                                         long FNAMESLength,
                                         long CWLength,
                                         long CUSERLength
                                         );

      [DllImport("libnag_nag.so", EntryPoint="e04vhf_")]
        public static extern void E04VHF_(
                                         ref int START, ref int NF, ref int N,
                                         ref int NXNAME, ref int NFNAME, ref double OBJADD,
                                         ref int OBJROW, string PROB, E04VHF_USRFUN_DELEGATE USRFUN,
                                         [In] int[] IAFUN, [In] int[] JAVAR, [In] double[] A,
                                         ref int LENA, ref int NEA, [In] int[] IGFUN,
                                         [In] int[] JGVAR, ref int LENG, ref int NEG,
                                         [In] double[] XLOW, [In] double[] XUPP, [In,Out] byte[,] XNAMES,
                                         [In] double[] FLOW, [In] double[] FUPP, [In,Out] byte[,] FNAMES,
                                         [In,Out] double[] X, [In,Out] int[] XSTATE, [Out] double[] XMUL,
                                         [In,Out] double[] F, [In,Out] int[] FSTATE, [In,Out] double[] FMUL,
                                         ref int NS, ref int NINF, ref double SINF,
                                         [In,Out] byte[,] CW, ref int LENCW, [In,Out] int[] IW,
                                         ref int LENIW, [In,Out] double[] RW, ref int LENRW,
                                         [In,Out] byte[,] CUSER, [In,Out] int[] IUSER, [In,Out] double[] RUSER,
                                         ref int IFAIL,
                                         long PROBLength,
                                         long XNAMESLength,
                                         long FNAMESLength,
                                         long CWLength,
                                         long CUSERLength
                                         );

      [DllImport("NLW6I29DE_nag.dll")] /* Import from DLL, the C# compiler provides a rudimentary check of the signature */
        public static extern void E04VGF(
                                         [In,Out] byte[,] CW,
                                         ref int LENCW,
                                         [Out] int[] IW,
                                         ref int LENIW,
                                         [Out] double[] RW,
                                         ref int LENRW,
                                         ref int IFAIL,
                                         long CWLength
                                         );

      [DllImport("libnag_nag.so", EntryPoint="e04vgf_")]
        public static extern void E04VGF_(
                                         [In,Out] byte[,] CW,
                                         ref int LENCW,
                                         [Out] int[] IW,
                                         ref int LENIW,
                                         [Out] double[] RW,
                                         ref int LENRW,
                                         ref int IFAIL,
                                         long CWLength
                                         );

      [DllImport("NLW6I29DE_nag.dll")] /* Import from DLL, the C# compiler provides a rudimentary check of the signature */
        public static extern void E04VMF(
                                         string ST,
                                         ref int IVALUE,
                                         [In,Out] byte[,] CW,
                                         [In,Out] int[] IW,
                                         [In,Out] double[] RW,
                                         ref int IFAIL,
                                         long STLength,
                                         long CWLength
                                         );

      [DllImport("libnag_nag.so", EntryPoint="e04vmf_")]
        public static extern void E04VMF_(
                                         string ST,
                                         ref int IVALUE,
                                         [In,Out] byte[,] CW,
                                         [In,Out] int[] IW,
                                         [In,Out] double[] RW,
                                         ref int IFAIL,
                                         long STLength,
                                         long CWLength
                                         );

        // Convenience routine copies a string into a byte array
        private static void setName(byte[,] name, int i, string sname)
        {
            for (int j = 0; j < sname.Length; j++)
                name[i, j] = (byte)sname[j];
        }

        public static void Main()
        {
            // Create delegate object
            E04VHF_USRFUN_DELEGATE userfun = new E04VHF_USRFUN_DELEGATE(usrfun);
            int n = 4;
            int nf = 6;
            int nea = 8;
            int neg = 8;
            int lena = nea;
            int leng = neg;
            int objrow = 6;
            int start = 0;
            int lencw = 600;
            int leniw = 600;
            int lenrw = 600;
            int char_array_elem_len = 8;

            byte[,] cuser = new byte[1, char_array_elem_len];
            int[] iuser = new int[1];
            double[] ruser = new double[1];

            byte[,] cw = new byte[lencw, char_array_elem_len];
            int[] iw = new int[leniw];
            double[] rw = new double[lenrw];

            int nxname = n;
            string[] xnames = { "X1      ", "X2      ", "X3      ", "X4      " };
            byte[,] xnames_byte = new byte[nxname, char_array_elem_len];
            for (int i = 0; i < nxname; i++)
            {
                setName(xnames_byte, i, xnames[i]);
            }

            int nfname = nf;
            string[] fnames = { "NlnCon 1", "NlnCon 2", "NlnCon 3", "NlnCon 4", "NlnCon 5", "NlnCon 6" };
            byte[,] fnames_byte = new byte[nfname, char_array_elem_len];
            for (int i = 0; i < nfname; i++)
            {
                setName(fnames_byte, i, fnames[i]);
            }

            int[] iafun = { 1, 2, 4, 4, 5, 5, 6, 6 };
            int[] javar = { 3, 4, 1, 2, 1, 2, 3, 4 };
            double[] a = { -1.0, -1.0, -1.0, 1.0, 1.0, -1.0, 3.0, 2.0 };

            int[] igfun = { 1, 1, 2, 2, 3, 3, 6, 6 };
            int[] jgvar = { 1, 2, 1, 2, 1, 2, 3, 4 };

            double[] xlow = { -0.55, -0.55, 0.0, 0.0 };
            double[] xupp = { 0.55, 0.55, 1200.0, 1200.0 };

            double[] flow = { -894.8, -894.8, -1294.8, -0.55, -0.55, -1.0e25 };
            double[] fupp = { -894.8, -894.8, -1294.8, 1.0e25, 1.0e25, 1.0e25 };

            double[] x = { 0.0, 0.0, 0.0, 0.0 };
            int[] xstate = { 0, 0, 0, 0 };
            double[] xmul = { 0.0, 0.0, 0.0, 0.0 };

            double[] f = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
            int[] fstate = { 0, 0, 0, 0, 0, 0 };
            double[] fmul = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };

            int ns = 0;
            int ninf = 0;
            double sinf = 0;
            double objadd = 0.0;
            int ifail = 1;

            if (RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
            {
                E04VGF(cw, ref lencw, iw, ref leniw, rw, ref lenrw, ref ifail, char_array_elem_len);
            }
            else if (RuntimeInformation.IsOSPlatform(OSPlatform.Linux))
            {
                E04VGF_(cw, ref lencw, iw, ref leniw, rw, ref lenrw, ref ifail, char_array_elem_len);
            }
            else
            {
                Console.WriteLine("Unsupported OS Platform");
                return;
            }
            if (ifail != 0)
            {
                Console.WriteLine("E04VMF: ifail = {0,3}", ifail);
                return;
            }         

            string option_str = "Print File";
            int nout = 6;
            string prob = "12345678";

            ifail = 1;
            if (RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
            {
                E04VMF(option_str, ref nout, cw, iw, rw, ref ifail, option_str.Length, char_array_elem_len);
                if (ifail != 0)
                {
                    Console.WriteLine("E04VMF: ifail = {0,3}", ifail);
                    return;
                }         

                ifail = 1;
                E04VHF(ref start, ref nf, ref n, ref nxname, ref nfname, ref objadd, ref objrow, prob,
                        userfun, iafun, javar, a, ref lena, ref nea, igfun, jgvar, ref leng, ref neg, xlow, xupp,
                       xnames_byte, flow, fupp, fnames_byte, x, xstate, xmul,
                       f, fstate, fmul, ref ns, ref ninf, ref sinf, cw, ref lencw,
                       iw, ref leniw, rw, ref lenrw, cuser, iuser, ruser, ref ifail,
                       prob.Length, char_array_elem_len, char_array_elem_len, char_array_elem_len, char_array_elem_len);
            }
            else if (RuntimeInformation.IsOSPlatform(OSPlatform.Linux))
            {
                E04VMF_(option_str, ref nout, cw, iw, rw, ref ifail, option_str.Length, char_array_elem_len);
                if (ifail != 0)
                {
                    Console.WriteLine("E04VMF: ifail = {0,3}", ifail);
                    return;
                }         

                ifail = 1;
                E04VHF_(ref start, ref nf, ref n, ref nxname, ref nfname, ref objadd, ref objrow, prob,
                        userfun, iafun, javar, a, ref lena, ref nea, igfun, jgvar, ref leng, ref neg, xlow, xupp,
                       xnames_byte, flow, fupp, fnames_byte, x, xstate, xmul,
                       f, fstate, fmul, ref ns, ref ninf, ref sinf, cw, ref lencw,
                       iw, ref leniw, rw, ref lenrw, cuser, iuser, ruser, ref ifail,
                       prob.Length, char_array_elem_len, char_array_elem_len, char_array_elem_len, char_array_elem_len);
            }

            Console.WriteLine("E04VHF: ifail = {0,3}", ifail);
            if (ifail == 0)
            {
                Console.WriteLine("Final Objective Value = {0, 11:f1}", f[objrow - 1]);
                for (int i = 0; i < n; i++)
                {
                    Console.Write("{0, 9:f2}{1}", x[i], (i + 1) % 7 == 0 || i + 1 == n ? "\n" : "");
                }
                Console.WriteLine("");
            }

        }

        public static void usrfun(ref int status, ref int n, IntPtr x_ptr, ref int needf, ref int nf,
            IntPtr f_ptr, ref int needg, ref int leng, IntPtr g_ptr, IntPtr cuser_ptr,
            IntPtr iuser_ptr, IntPtr ruser_ptr, long char_array_elem_len)
        {
            double[] x = new double[n];
            Marshal.Copy(x_ptr, x, 0, n);

            double[] g = new double[leng];
            Marshal.Copy(g_ptr, g, 0, leng);

            double[] f = new double[nf];
            Marshal.Copy(f_ptr, f, 0, nf);
            if (needf > 0)
            {


                //         The nonlinear components of f_i(x) need to be assigned,
                //         for i = 1 to NF
                f[0] = 1000.00e+0 * Math.Sin(-x[0] - 0.250e+0) + 1000.00e+0 * Math.Sin(-x[1] - 0.250e+0);
                f[1] = 1000.00e+0 * Math.Sin(x[0] - 0.250e+0) + 1000.00e+0 * Math.Sin(x[0] - x[1] - 0.250e+0);
                f[2] = 1000.00e+0 * Math.Sin(x[1] - x[0] - 0.250e+0) + 1000.00e+0 * Math.Sin(x[1] - 0.250e+0);
                //         N.B. in this example there is no need to assign for the wholly
                // linear components f_4(x) and f_5(x).
                f[5] = 1.00e-6 * Math.Pow(x[2], 3) + 2.00e-6 * Math.Pow(x[3], 3) / 3.00e+0;
            }
            //
            if (needg > 0)
            {
                //         The derivatives of the function f_i(x) need to be assigned.
                //         G(k) should be set to partial derivative df_i(x)/dx_j where
                //         i = IGFUN(k) and j = IGVAR(k), for k = 1 to LENG.
                g[0] = -1000.00e+0 * Math.Cos(-x[0] - 0.250e+0);
                g[1] = -1000.00e+0 * Math.Cos(-x[1] - 0.250e+0);
                g[2] = 1000.00e+0 * Math.Cos(x[0] - 0.250e+0) + 1000.00e+0 * Math.Cos(x[0] - x[1] - 0.250e+0);
                g[3] = -1000.00e+0 * Math.Cos(x[0] - x[1] - 0.250e+0);
                g[4] = -1000.00e+0 * Math.Cos(x[1] - x[0] - 0.250e+0);
                g[5] = 1000.00e+0 * Math.Cos(x[1] - x[0] - 0.250e+0) + 1000.00e+0 * Math.Cos(x[1] - 0.250e+0);
                g[6] = 3.00e-6 * (x[2]) * (x[2]);
                g[7] = 2.00e-6 * (x[3]) * (x[3]);
            }
            Marshal.Copy(g, 0, g_ptr, leng);
            Marshal.Copy(f, 0, f_ptr, nf);

        } // End of usrfun
    }
}
